import re
import requests
import math
import random
import numpy as np

def routes_to_solution(routes):
    """
    Converts a list of routes into a solution (list of routes).

    Args:
        routes (list): List of routes, where each route is a list of nodes.

    Returns:
        list: Solution (list of routes).
    """
    solution = []
    for i, route in enumerate(routes):
        assert route[0] == 0
        assert route[-1] == 0
        # Skip depot node for all routes except the first one
        if i > 0:
            route = route[1:]
        solution += route
    return solution

def add_depot_to_routes(routes):
    """
    Adds the depot node to the beginning and end of each route.

    Args:
        routes (list): List of routes, where each route is a list of nodes.

    Returns:
        list: List of routes, where each route is a list of nodes, with the depot added.

    Raises:
        ValueError: If routes is not a list or a nested list.
    """
    if not isinstance(routes, list):
        raise ValueError("Routes must be a list")

    for route in routes:
        if not isinstance(route, list):
            raise ValueError("Routes must be a nested list")

        route.insert(0, 0)
        route.append(0)

    return routes

def solution_to_routes(solution, partial=False):
    """
    Converts a solution (list of nodes) into a list of routes.

    Args:
        solution (list): Solution (list of nodes).

    Returns:
        list: List of routes, where each route is a list of nodes.
    """
    routes = []
    route = []
    for node in solution:
        if node == 0 and route:
            route.append(0)
            routes.append(route)
            route = []
        route.append(node)
    if partial:
        routes.append(route)
    return routes

def solution_to_adjacency_matrix(cvrp_solution):
    """
    Converts a CVRP solution into a symmetric adjacency matrix.

    Args:
        cvrp_solution (list): CVRP solution (list of nodes).

    Returns:
        np.ndarray: Symmetric adjacency matrix.
    """
    num_nodes = max(cvrp_solution) + 1
    matrix = np.zeros((num_nodes, num_nodes))

    for i in range(len(cvrp_solution) - 1):
        matrix[cvrp_solution[i]][cvrp_solution[i + 1]] = 1
        matrix[cvrp_solution[i + 1]][cvrp_solution[i]] = 1
    return matrix

def column_normalize_adjacency_matrix(adj_matrix):
    """
    Normalize an adjacency matrix by column.

    Args:
        adj_matrix (np.ndarray): Binary adjacency matrix.

    Returns:
        np.ndarray: Normalized adjacency matrix.
    """
    col_sums = np.sum(adj_matrix, axis=0)
    normalized_matrix = adj_matrix / col_sums[np.newaxis, :]
    return normalized_matrix

def parse_vrplib_file(source):
    """
    Parses a VRPLIB file and returns a dictionary with the following keys:
    - name: Name of the problem.
    - comment: Comment of the problem.
    - type: Type of the problem.
    - dimension: Number of nodes.
    - edge_weight_type: Type of edge weight.
    - capacity: Capacity of the vehicles.
    - node_coords: Dict of node coordinates.
    - demand: Dict of node demands.
    - depot_ids: List of depot IDs.

    Args:
        source (str): URL or path to the VRPLIB file.

    Returns:
        dict: Dictionary with the parsed data.
    """
    if source.startswith('http://') or source.startswith('https://'):
        response = requests.get(source)
        content = response.text
    else:
        with open(source, 'r') as file:
            content = file.read()

    data = {}

    # Extracting problem information using regex
    name_match = re.search(r'NAME\s*:\s*(\S+)', content)
    if name_match:
        data['name'] = name_match.group(1)

    comment_match = re.search(r'COMMENT\s*:\s*"(.+)"', content)
    if comment_match:
        data['comment'] = comment_match.group(1)

    type_match = re.search(r'TYPE\s*:\s*(\S+)', content)
    if type_match:
        data['type'] = type_match.group(1)

    dimension_match = re.search(r'DIMENSION\s*:\s*(\d+)', content)
    if dimension_match:
        data['dimension'] = int(dimension_match.group(1))

    edge_weight_type_match = re.search(r'EDGE_WEIGHT_TYPE\s*:\s*(\S+)', content)
    if edge_weight_type_match:
        data['edge_weight_type'] = edge_weight_type_match.group(1)

    capacity_match = re.search(r'CAPACITY\s*:\s*(\d+)', content)
    if capacity_match:
        data['capacity'] = int(capacity_match.group(1))

    node_coords_match = re.search(r'NODE_COORD_SECTION\s*(.+?)\s*DEMAND_SECTION', content, re.DOTALL)
    if node_coords_match:
        node_coords_str = node_coords_match.group(1).strip()
        node_coords = []
        for line in node_coords_str.split('\n'):
            node_id, x, y = re.findall(r'(\d+)\s+([-+]?\d*\.?\d+)\s+([-+]?\d*\.?\d+)', line)[0]
            node_coords.append([float(x), float(y)])
        data['node_coords'] = np.array(node_coords)

    # Extracting demand information
    demand_section_match = re.search(r'DEMAND_SECTION\s*(.+?)\s*DEPOT_SECTION', content, re.DOTALL)
    if demand_section_match:
        demand_section_str = demand_section_match.group(1).strip()
        demand = []
        for line in demand_section_str.split('\n'):
            node_id, demand_val = re.findall(r'(\d+)\s+(-?\d+)', line)[0]
            demand.append(int(demand_val))
        data['demand'] = np.array(demand)

    # Extracting depot information
    depot_section_match = re.search(r'DEPOT_SECTION\s*(.+?)\s*EOF', content, re.DOTALL)
    if depot_section_match:
        depot_section_str = depot_section_match.group(1).strip()
        depot_ids = [int(node_id) for node_id in re.findall(r'(\d+)\s*[^-0-9]', depot_section_str)]
        data['depot_ids'] = depot_ids

    return data

def parse_vrplib_solution(source):
    """
    Parses a solution file from the VRPLIB website and returns the routes and cost.

    Args:
        source (str): Path to the solution file or URL to the solution file.

    Returns:
        tuple: Tuple containing the routes (list of lists) and the cost (float).
    """
    route_regex = re.compile(r'Route #(\d+): (.+)')
    cost_regex = re.compile(r'Cost ([\d.]+)')

    if source.startswith('http://') or source.startswith('https://'):
        response = requests.get(source)
        solution_text = response.text
    else:
        with open(source, 'r') as file:
            solution_text = file.read()

    routes = []
    cost = None

    lines = solution_text.strip().split('\n')
    for line in lines:
        route_match = route_regex.match(line)
        if route_match:
            route_number = int(route_match.group(1))
            nodes = list(map(int, route_match.group(2).split()))
            routes.append(nodes)

        cost_match = cost_regex.match(line)
        if cost_match:
            cost = float(cost_match.group(1))

    return routes, cost

def get_cvrp_cost(routes_or_solution, coordinates, uchoa=False):
    """
    Compute the total cost of a CVRP solution.

    Args:
        routes_or_solution (list or solution): List of routes or a solution.
        coordinates (np.array): Array of node coordinates with shape (n_nodes, 2).
        uchoa (bool): Whether to round the distances to the nearest integer. Follows the 2014 paper.

    Returns:
        float: Total cost of the CVRP solution.
    """
    if isinstance(routes_or_solution, list) and isinstance(routes_or_solution[0], list):
        solution = routes_to_solution(routes_or_solution)
    elif isinstance(routes_or_solution, list):
        solution = routes_or_solution
    else:
        raise ValueError("Invalid input. Expected a nested list of routes or a solution.")

    total_cost = 0.0
    for i in range(len(solution) - 1):
        node1 = solution[i]
        node2 = solution[i + 1]
        coord1 = coordinates[node1]
        coord2 = coordinates[node2]
        distance = math.dist(coord1, coord2)
        if uchoa:
            distance = round(distance)
        total_cost += distance

    return total_cost

def get_all_route_demands(routes, demand):
    """
    Compute the demand of each route.

    Args:
        routes (list): List of routes.

    Returns:
        route_demand (list): List containing the demand of each route.
    """
    route_demands = []
    for route in routes:
        route_demand = 0
        for node in route:
            try:
                route_demand += demand[node]
            except KeyError:
                raise Exception(f"Node {node} not found in the demand dictionary")
        route_demands.append(route_demand)
    return route_demands

def get_route_demand(route, demands):
    """
    Compute the demand of a route.

    Args:
        route (list): List of nodes in the route.
        demands (list): List containing the corresponding demands for each node.

    Returns:
        int: Demand of the route.
    """
    route_demand = sum(map(lambda node: demands[node], route))
    return route_demand

def is_feasible(routes, demand, capacity):
    """
    Check if a CVRP solution is feasible.

    Args:
        routes (list): List of routes.
        demand (dict): Dictionary containing node IDs as keys and corresponding demand as values.
        capacity (int): Vehicle capacity.

    Returns:
        bool: True if the solution is feasible, False otherwise.
    """
    # Convert routes to a single solution
    solution = routes_to_solution(routes)

    # Check if each node appears only once
    if set(solution) != set(demand.keys()):
        return False

    # Compute the demand for each route
    for route in routes:
        total_demand = sum(demand[node] for node in route)
        if total_demand > capacity:
            return False

    return True

def get_cvrp_problem(num_nodes):
    """
    Generate a random CVRP problem.
    Follows: "https://arxiv.org/pdf/1802.04240.pdf"

    Args:
        num_nodes (int): Number of nodes in the problem. Excluding depot.

    Returns:
        tuple: Tuple containing the node coordinates (np.ndarray) and the demand (np.ndarray).
    """
    # Generate depot
    depot_id = 0
    depot_coords = np.array([random.uniform(0, 1), random.uniform(0, 1)])
    depot_demand = np.array([0])

    # Generate node coordinates array
    node_coords = np.empty((num_nodes + 1, 2))
    node_coords[depot_id] = depot_coords

    # Generate demand array
    demands = np.empty((num_nodes + 1))
    demands[depot_id] = depot_demand

    # Generate node coordinates and demand
    for node_id in range(1, num_nodes + 1):
        x = random.uniform(0, 1)
        y = random.uniform(0, 1)
        demand_val = random.randint(1, 9)
        node_coords[node_id] = np.array([float(x), float(y)])
        demands[node_id] = int(demand_val)

    return node_coords, demands

def pad_matrix(matrix, new_shape, constant_value=0):
    """
    Pads a matrix to the right and bottom with zeros.

    Args:
        matrix: the input matrix to pad.
        new_shape: the desired shape after padding.

    Returns:
        returns the padded matrix.

    ```python
    import numpy as np
    matrix = np.arange(10).reshape(2, 5)
    pad_matrix(matrix, (3, 6))
    array([[0, 1, 2, 3, 4, 0],
           [5, 6, 7, 8, 9, 0],
           [0, 0, 0, 0, 0, 0]])
    ```
    """
    assert len(new_shape) == len(matrix.shape), "new_shape and matrix dimensions must match"
    padding = [(0, new_dim - old_dim) if new_dim > old_dim else (0, 0) for old_dim, new_dim in zip(matrix.shape, new_shape)]
    return np.pad(matrix, padding, 'constant', constant_values=constant_value)

def get_submatrix(indices, matrix, new_shape):
    """
    Gets a submatrix from a matrix and pads it to the right and bottom with zeros.

    Args:
        indices: the indices to index on the matrix.
        matrix: the input matrix to index.
        new_shape: the desired shape after padding.

    Returns:
        returns the submatrix after indexing and padding.

    ```python
    import numpy as np
    matrix = np.arange(10).reshape(2, 5)
    indices = [1, 2, 3]
    array([[1, 2, 3, 0, 0, 0],
           [6, 7, 8, 0, 0, 0],
           [0, 0, 0, 0, 0, 0]])
    ```
    """
    submatrix = matrix[:, indices]
    return pad_matrix(submatrix, new_shape)

def get_random_solution(N):
    """
    Generates a random CVRP solution. Mostly for debugging purposes.

    Args:
        N (int): Number of nodes in the problem. Excluding depot.
    
    Returns:
        lst (list): List containing the solution.
    """
    # Generate a random list from 0 to N
    lst = list(range(N + 1))

    # Insert 0 at the end
    lst.append(0)

    # Randomly insert 1 to 5 zeros in between
    num_zeros = random.randint(1, 5)
    for _ in range(num_zeros):
        index = random.randint(1, len(lst) - 1)  # Select a random index between 1 and len(lst)-1
        while lst[index] == 0 or lst[index-1] == 0:
            index = random.randint(1, len(lst) - 1)  # Select a new index if adjacent elements are already 0
        lst.insert(index, 0)

    return lst
    
def get_logit_mask(sol, demands, capacity):
    """
    Gets a logit mask for a given solution. Note that this only takes a full solution.

    Args:
        sol (list): the solution to mask.
        demands (np.ndarray): the demands of each node.
        capacity (int): the capacity of the vehicle.
    
    Returns:
        mask (np.ndarray): the logit mask.
    """
    prob = set(sol)
    mask = np.full((len(sol), len(prob)), 0.0) # mark all as valid
    for i in range(len(sol)): # NOTE this loop is 100% correct, don't change this.
        # a is all nodes that I visited
        a = set(sol[:i+1])
        if sol[i] != 0:
            a.remove(0)
        # mark all visited nodes as invalid
        mask[i, list(a)] = float("-inf")

        # mark all infeasible demand nodes as invalid
        # this crazy implementation is due to cProfile saying 
        # list comprehension is super slow. numpy is fast.
        c_nodes = list(set(prob) - set(sol[:i+1]) - set([0]))
        last_route = solution_to_routes(sol[:i+1], partial=True)[-1]
        cur_demand = get_route_demand(last_route, demands)
        c_demands = demands[c_nodes]
        next_demands = c_demands + cur_demand
        sel = np.squeeze(np.argwhere(next_demands > capacity))
        infeasible_nodes = np.array(c_nodes)[sel]
        if a == set(prob):
            mask[i, :] = 0
        else:
            mask[i, infeasible_nodes.tolist()] = float("-inf")
    return mask




import unittest
from routeml.utils import get_cvrp_cost
import math

class TestGetCVRPCost(unittest.TestCase):
    def test_get_cvrp_cost_with_routes(self):
        routes = [[0, 1, 2, 0], [0, 3, 4, 5, 0]]
        coordinates = {0: (0, 0), 1: (1, 1), 2: (2, 2), 3: (3, 3), 4: (4, 4), 5: (5, 5)}

        cost = get_cvrp_cost(routes, coordinates)
        self.assertAlmostEqual(cost, 2 * math.sqrt(50) + 2 * math.sqrt(8))

    def test_get_cvrp_cost_with_solution(self):
        solution = [0, 1, 2, 3, 4, 5, 0]
        coordinates = {0: (0, 0), 1: (1, 1), 2: (2, 2), 3: (3, 3), 4: (4, 4), 5: (5, 5)}

        cost = get_cvrp_cost(solution, coordinates)
        self.assertAlmostEqual(cost, 2 * math.sqrt(50))

    def test_get_cvrp_cost_invalid_input(self):
        invalid_input = "invalid"
        coordinates = {0: (0, 0), 1: (1, 1), 2: (2, 2)}

        with self.assertRaises(ValueError):
            get_cvrp_cost(invalid_input, coordinates)

if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
from routeml.utils import pad_matrix, get_submatrix

class TestMatrixOperations(unittest.TestCase):
    def test_pad_matrix(self):
        # Test 1: Basic padding
        matrix = np.ones((5, 7))
        new_shape = (10, 9)
        padded_matrix = pad_matrix(matrix, new_shape)
        self.assertEqual(padded_matrix.shape, new_shape)
        self.assertEqual(np.sum(padded_matrix), 35)  # Original sum

        # Test 2: No padding needed
        matrix = np.ones((5, 7))
        new_shape = (5, 7)
        padded_matrix = pad_matrix(matrix, new_shape)
        np.testing.assert_array_equal(padded_matrix, matrix)

        # Test 3: Padding only one dimension
        matrix = np.ones((5, 7))
        new_shape = (5, 9)
        padded_matrix = pad_matrix(matrix, new_shape)
        self.assertEqual(padded_matrix.shape, new_shape)
        self.assertEqual(np.sum(padded_matrix), 35)  # Original sum

    def test_get_submatrix(self):
        matrix = np.arange(500).reshape((10, 50))  # 10x50 matrix
        indices = [1, 3, 5, 7, 9]
        submatrix = get_submatrix(indices, matrix, (10, 100))
        for i, idx in enumerate(indices):
            np.testing.assert_array_equal(submatrix[:, i], matrix[:, idx])

    def test_pad_and_get_submatrix(self):
        N = 1000
        M = 1200
        l_matrix = np.random.rand(7, N)
        l_matrix = pad_matrix(l_matrix, (9, N))
        indices = [1, 3, 5, 7, 9]
        r_matrix = get_submatrix(indices, l_matrix, (9, M))
        for i, idx in enumerate(indices):
            np.testing.assert_array_equal(r_matrix[:, i], l_matrix[:, idx])
        self.assertEqual(np.sum(l_matrix[:, indices]), np.sum(r_matrix[:, :len(indices)]))

if __name__ == '__main__':
    unittest.main()



import unittest
import requests
import os
import tempfile
from routeml.utils import parse_vrplib_file


class VRPLIBParserTestCase(unittest.TestCase):
    def test_parse_vrplib_file(self):
        url = "http://vrp.atd-lab.inf.puc-rio.br/media/com_vrp/instances/X/X-n101-k25.vrp"
        response = requests.get(url)
        content = response.text

        # Create a temporary directory
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = os.path.join(temp_dir, "X-n101-k25.vrp")

            # Write the file content to the temporary directory
            with open(file_path, "w") as file:
                file.write(content)

            # Perform the parsing using the local file
            parsed_data = parse_vrplib_file(file_path)

            # Perform assertions on the parsed data
            self.assertEqual(parsed_data['name'], "X-n101-k25")
            self.assertEqual(parsed_data['type'], "CVRP")
            self.assertEqual(parsed_data['dimension'], 101)
            self.assertEqual(parsed_data['edge_weight_type'], "EUC_2D")
            self.assertEqual(parsed_data['capacity'], 206)

            self.assertEqual(parsed_data['node_coords'].shape[0], 101)
            self.assertEqual(parsed_data['node_coords'].shape[1], 2)
            self.assertEqual(parsed_data['demand'].shape[0], 101)
            self.assertEqual(len(parsed_data['depot_ids']), 1)

            # Additional assertions or validations on the parsed data

            print("VRPLIB file parsed successfully!")


if __name__ == '__main__':
    unittest.main()



import unittest
from routeml.utils import get_all_route_demands, get_route_demand

class TestGetRouteDemand(unittest.TestCase):
    def test_get_route_demand(self):
        routes = [[0, 1, 2, 0], [0, 3, 4, 0], [0, 5, 6, 7, 0]]
        demand = {0: 0, 1: 10, 2: 20, 3: 15, 4: 25, 5: 30, 6: 5, 7: 8}

        total_demand = get_all_route_demands(routes, demand)
        self.assertEqual(total_demand, [30, 40, 43])

    def test_get_route_demand_empty_routes(self):
        routes = []
        demand = {0: 0, 1: 10, 2: 20}

        total_demand = get_all_route_demands(routes, demand)
        self.assertEqual(total_demand, [])

    def test_get_route_demand_empty_demand(self):
        routes = [[0, 1, 2, 0], [0, 3, 4, 0]]
        demand = {}

        with self.assertRaises(Exception) as context:
            total_demand = get_all_route_demands(routes, demand)
        
        self.assertEqual(str(context.exception), "Node 0 not found in the demand dictionary")

    def test_get_route_demand_zero_demand(self):
        routes = [[0, 1, 2, 0], [0, 3, 4, 0]]
        demand = {0: 0, 1: 0, 2: 0, 3: 0, 4: 0}

        total_demand = get_all_route_demands(routes, demand)
        self.assertEqual(total_demand, [0, 0])

    def test_get_route_demand_with_missing_node(self):
        routes = [[1, 2, 3], [4, 5, 6]]
        demand = {1: 10, 2: 20, 4: 15, 6: 30}
        
        with self.assertRaises(Exception) as context:
            total_demand = get_all_route_demands(routes, demand)
        
        self.assertEqual(str(context.exception), "Node 3 not found in the demand dictionary")

    def test_empty_route(self):
        route = []
        demands = []
        self.assertEqual(get_route_demand(route, demands), 0)

    def test_non_empty_route_positive_demands(self):
        route = [0, 1, 2, 3, 4, 5]
        demands = [0, 10, 5, 8, 12, 3]
        self.assertEqual(get_route_demand(route, demands), 38)

    def test_non_empty_route_zero_demands(self):
        route = [0, 1, 2]
        demands = [0, 0, 0]
        self.assertEqual(get_route_demand(route, demands), 0)

    def test_route_with_duplicate_nodes(self):
        route = [0, 1, 2, 3, 2, 4, 1]
        demands = [0, 5, 3, 2, 4, 6]
        self.assertEqual(get_route_demand(route, demands), 22)

    def test_route_with_negative_demands(self):
        route = [0, 4, 3, 2, 1]
        demands = [0, -10, -5, -8, -12]
        self.assertEqual(get_route_demand(route, demands), -35)

if __name__ == '__main__':
    unittest.main()



import unittest
from routeml.utils import routes_to_solution

class TestRoutesToSolution(unittest.TestCase):
    def test_routes_to_solution(self):
        routes = [[0, 1, 2, 3, 0], [0, 4, 5, 6, 0]]
        expected_solution = [0, 1, 2, 3, 0, 4, 5, 6, 0]
        solution = routes_to_solution(routes)
        self.assertEqual(solution, expected_solution)

    def test_routes_to_solution_empty_routes(self):
        routes = []
        expected_solution = []
        solution = routes_to_solution(routes)
        self.assertEqual(solution, expected_solution)

    def test_routes_to_solution_single_route(self):
        routes = [[0, 1, 2, 3, 0]]
        expected_solution = [0, 1, 2, 3, 0]
        solution = routes_to_solution(routes)
        self.assertEqual(solution, expected_solution)

    def test_routes_to_solution_multiple_routes(self):
        routes = [[0, 1, 2, 3, 0], [0, 4, 5, 6, 0], [0, 7, 8, 9, 0]]
        expected_solution = [0, 1, 2, 3, 0, 4, 5, 6, 0, 7, 8, 9, 0]
        solution = routes_to_solution(routes)
        self.assertEqual(solution, expected_solution)

if __name__ == '__main__':
    unittest.main()



import unittest
from routeml.utils import solution_to_routes

class TestSolutionToRoutes(unittest.TestCase):
    def test_solution_to_routes(self):
        solution = [0, 1, 2, 3, 0, 4, 5, 0]
        expected_routes = [[0, 1, 2, 3, 0], [0, 4, 5, 0]]
        routes = solution_to_routes(solution)
        self.assertEqual(routes, expected_routes)

    def test_solution_to_routes_empty_solution(self):
        solution = []
        expected_routes = []
        routes = solution_to_routes(solution)
        self.assertEqual(routes, expected_routes)

    def test_solution_to_routes_single_route(self):
        solution = [0, 1, 2, 3, 4, 0]
        expected_routes = [[0, 1, 2, 3, 4, 0]]
        routes = solution_to_routes(solution)
        self.assertEqual(routes, expected_routes)

    def test_solution_to_routes_multiple_routes(self):
        solution = [0, 1, 0, 2, 3, 0, 4, 0]
        expected_routes = [[0, 1, 0], [0, 2, 3, 0], [0, 4, 0]]
        routes = solution_to_routes(solution)
        self.assertEqual(routes, expected_routes)

    def test_partial_solution_to_routes_1(self):
        solution = [0, 1, 0, 2, 3, 0, 4]
        expected_routes = [[0, 1, 0], [0, 2, 3, 0], [0, 4]]
        routes = solution_to_routes(solution, partial=True)
        self.assertEqual(routes, expected_routes)

    def test_partial_solution_to_routes_2(self):
        solution = [0, 1, 0, 2, 3, 0, 4, 5]
        expected_routes = [[0, 1, 0], [0, 2, 3, 0], [0, 4, 5]]
        routes = solution_to_routes(solution, partial=True)
        self.assertEqual(routes, expected_routes)

if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
import matplotlib.pyplot as plt
from routeml.utils import get_cvrp_problem
from routeml.draw import plot_routes
from routeml.solvers import hgs_solve

class CVRPIntegrationTest(unittest.TestCase):
    def test_solve_and_plot_cvrp(self):
        # Generate a CVRP problem
        num_nodes = 100
        node_coords, demands = get_cvrp_problem(num_nodes)

        # Solve the problem using HGS or any other solver
        result = hgs_solve(node_coords, demands, 50, time_limit=2)
        # (Replace with your solver implementation)

        # Obtain the solution routes
        routes = result.routes

        # Plot and save the solution as a PNG file
        plot_routes(routes, node_coords, save_path="test_output/solution.png")

        # Add assertions if needed
        # self.assertEqual(...)

if __name__ == '__main__':
    unittest.main()


import matplotlib.pyplot as plt

def plot_routes(routes, node_coords, save_path):
    """
    Plot the routes on a 2D plane.

    Args:
        routes (list): A list of routes, where each route is a list of node IDs.
        node_coords (dict): A dictionary of node coordinates, where the key is
            the node ID and the value is a tuple of the x and y coordinates.
        save_path (str): The path to save the plot to.

    Returns:
        None
    """
    # Create a list of unique colors for each route
    colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k']

    # Plot each route with a different color
    for i, route in enumerate(routes):
        x = [node_coords[node][0] for node in route]
        y = [node_coords[node][1] for node in route]
        plt.plot(x, y, 'o-', color=colors[i % len(colors)])

    # Plot the depot node with an X
    depot_x, depot_y = node_coords[0]
    plt.plot(depot_x, depot_y, 'rx', markersize=10, label='Depot')

    plt.xlabel('X')
    plt.ylabel('Y')
    plt.title('Routes')
    plt.savefig(save_path)
    plt.close()



import unittest
import numpy as np
from routeml.utils import solution_to_adjacency_matrix, get_random_solution, column_normalize_adjacency_matrix

class TestSolutionToAdjacencyMatrix(unittest.TestCase):
    def test_cvrp_solution_to_adjacency_matrix(self):
        cvrp_solution = [0, 1, 2, 3, 4, 0, 5, 6, 7, 0]
        expected_matrix = np.array([[0, 1, 0, 0, 1, 1, 0, 1],
                                    [1, 0, 1, 0, 0, 0, 0, 0],
                                    [0, 1, 0, 1, 0, 0, 0, 0],
                                    [0, 0, 1, 0, 1, 0, 0, 0],
                                    [1, 0, 0, 1, 0, 0, 0, 0],
                                    [1, 0, 0, 0, 0, 0, 1, 0],
                                    [0, 0, 0, 0, 0, 1, 0, 1],
                                    [1, 0, 0, 0, 0, 0, 1, 0]])

        result = solution_to_adjacency_matrix(cvrp_solution)
        np.testing.assert_array_equal(result, expected_matrix)

    def test_check_all_rows_sum_to_2(self):
        for i in range(10):
            cvrp_solution = get_random_solution(50)
            result = solution_to_adjacency_matrix(cvrp_solution)
            num_zeros = np.sum(np.array(cvrp_solution) == 0)
            # Check if all rows except the first row sum to 2
            for i in range(1, result.shape[0]):
                row_sum = np.sum(result[i])
                self.assertIn(row_sum, [1, 2], f"Row {i} does not sum to 1 or 2.")

    def test_column_normalize_adj_mat(self):
        cvrp_solution = get_random_solution(50)
        result = solution_to_adjacency_matrix(cvrp_solution)
        result = column_normalize_adjacency_matrix(result)
        np.testing.assert_array_almost_equal(np.sum(result, axis=0), np.ones(result.shape[0]))

if __name__ == '__main__':
    unittest.main()



import unittest
import requests
from routeml.utils import parse_vrplib_file, parse_vrplib_solution, get_cvrp_cost, add_depot_to_routes

class VRPLIBIntegrationTestCase(unittest.TestCase):
    def test_vrplib_integration(self):
        files = [
            [
                "http://vrp.atd-lab.inf.puc-rio.br/media/com_vrp/instances/X/X-n134-k13.vrp",
                "http://vrp.atd-lab.inf.puc-rio.br/media/com_vrp/instances/X/X-n134-k13.sol",
                True
            ],
            [
                "http://vrp.atd-lab.inf.puc-rio.br/media/com_vrp/instances/X/X-n101-k25.vrp",
                "http://vrp.atd-lab.inf.puc-rio.br/media/com_vrp/instances/X/X-n101-k25.sol",
                True
            ],
            [
                "http://vrp.atd-lab.inf.puc-rio.br/media/com_vrp/instances/Li/Li_21.vrp",
                "http://vrp.atd-lab.inf.puc-rio.br/media/com_vrp/instances/Li/Li_21.sol",
                False
            ]
        ]
        for z in files:
            vrp_file_url, vrp_sol_url, uchoa = z
            vrplib_data = parse_vrplib_file(vrp_file_url)
            vrplib_solution, vrplib_cost = parse_vrplib_solution(vrp_sol_url)
            vrplib_solution = add_depot_to_routes(vrplib_solution)

            # 3. Get cost from parsed VRPLIB file
            cost = get_cvrp_cost(vrplib_solution, vrplib_data['node_coords'], uchoa=uchoa)

            # 4. Assert that the cost matches the VRPLIB solution
            self.assertAlmostEqual(vrplib_cost, cost, places=3)


if __name__ == '__main__':
    unittest.main()



import unittest
from routeml.utils import is_feasible

class TestIsFeasible(unittest.TestCase):
    def test_is_feasible(self):
        demand = {0: 0, 1: 10, 2: 15, 3: 8, 4: 12}
        capacity = 20

        routes1 = [[0, 1, 2, 3, 0], [0, 4, 0]]
        self.assertFalse(is_feasible(routes1, demand, capacity))

        routes2 = [[0, 1, 0], [0, 2, 0], [0, 3, 0], [0, 4, 0]]
        self.assertTrue(is_feasible(routes2, demand, capacity))

        routes3 = [[0, 1, 2, 4, 0], [0, 3, 0]]
        self.assertFalse(is_feasible(routes3, demand, capacity))

        routes4 = [[0, 1, 3, 0], [0, 2, 4, 0]]
        self.assertFalse(is_feasible(routes4, demand, capacity))

        routes5 = []
        self.assertFalse(is_feasible(routes5, demand, capacity))

        # Single route exceeding capacity
        routes6 = [[0, 1, 2, 3, 4, 0]]
        self.assertFalse(is_feasible(routes6, demand, capacity))

        # Duplicate nodes in a single route
        routes7 = [[0, 1, 2, 2, 3, 0], [0, 4, 0]]
        self.assertFalse(is_feasible(routes7, demand, capacity))

        # Duplicate nodes across routes
        routes8 = [[0, 1, 2, 3, 0], [0, 3, 4, 0]]
        self.assertFalse(is_feasible(routes8, demand, capacity))

        # Nodes not present in demand dictionary
        routes9 = [[0, 1, 2, 5, 0], [0, 3, 4, 0]]
        self.assertFalse(is_feasible(routes9, demand, capacity))

if __name__ == '__main__':
    unittest.main()



import unittest
import requests
import tempfile
import os
from routeml.utils import parse_vrplib_solution


class VRPSolutionParserTestCase(unittest.TestCase):
    def test_parse_solution_from_url(self):
        solution_url = 'http://vrp.atd-lab.inf.puc-rio.br/media/com_vrp/instances/Li/Li_22.sol'
        response = requests.get(solution_url)
        solution_text = response.text
        # Create a temporary directory
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = os.path.join(temp_dir, "Li_22.sol")

            # Write the file content to the temporary directory
            with open(file_path, "w") as file:
                file.write(solution_text)

            routes, cost = parse_vrplib_solution(file_path)

            # Perform assertions on the parsed routes and cost
            self.assertEqual(len(routes), 15)
            self.assertEqual(cost, 14499.04468)

            # Additional assertions or validations on the parsed data

            print("VRP solution parsed successfully!")


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

setup(
    name='routeml',
    version='0.8.0',
    description='Python package for CVRP utilities',
    author='Your Name',
    author_email='jkschin@mit.edu',
    url='https://github.com/jkschin/routeml',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    install_requires=[
        "numpy",
        "scipy",
        "matplotlib",
        "hygese"
        # Add any dependencies required by your package
    ],
)



import numpy as np
import hygese as hgs
from scipy.spatial import distance_matrix

def hgs_solve(coords, demands, capacity, time_limit=5):
    """
    coords: a numpy array of coordinates, where each row represents the x and y coordinates of a node
    demands: a numpy array of demands, where index 0 represents the demand of the depot (0)
    capacity: integer, capacity of all vehicles
    time_limit: integer, maximum time limit for the solver in seconds
    """

    n = len(coords)
    x = coords[:, 0]
    y = coords[:, 1]

    # Solver initialization
    ap = hgs.AlgorithmParameters(timeLimit=time_limit)
    hgs_solver = hgs.Solver(parameters=ap, verbose=False)

    # Data preparation
    data = dict()
    data['x_coordinates'] = x
    data['y_coordinates'] = y
    data['distance_matrix'] = distance_matrix(coords, coords)
    data['service_times'] = np.zeros(n)
    data['demands'] = demands
    data['vehicle_capacity'] = capacity
    data['num_vehicles'] = 100  # Update with the appropriate number of vehicles
    data['depot'] = 0

    result = hgs_solver.solve_cvrp(data)
    return result



import unittest
import numpy as np
from routeml.utils import get_logit_mask

class TestGetLogitMask(unittest.TestCase):
    def test_get_logit_mask_1(self):
        sol = [0, 1, 2, 0, 3, 4]
        demands = np.array([0, 5, 8, 3, 6])
        capacity = 15
        expected_mask = np.array([[-np.inf, 0., 0., 0., 0.],
                [0., -np.inf, 0., 0., 0.],
                [0., -np.inf, -np.inf, -np.inf, -np.inf],
                [-np.inf, -np.inf, -np.inf, 0., 0.],
                [0., -np.inf, -np.inf, -np.inf, 0.],
                [0., -np.inf, -np.inf, -np.inf, -np.inf]])

        result_mask = get_logit_mask(sol, demands, capacity)

        np.testing.assert_array_equal(result_mask, expected_mask)

if __name__ == '__main__':
    unittest.main()
